package pucrs.hslcommunicationservice.service;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageRequest;
import org.springframework.stereotype.Service;
import pucrs.hslcommunicationservice.aws.S3Client;
import pucrs.hslcommunicationservice.aws.SNSClient;
import pucrs.hslcommunicationservice.configuration.security.TokenAuthenticationService;
import pucrs.hslcommunicationservice.domain.Announcement;
import pucrs.hslcommunicationservice.domain.Attachment;
import pucrs.hslcommunicationservice.domain.User;
import pucrs.hslcommunicationservice.domain.UserAnnouncementView;
import pucrs.hslcommunicationservice.repository.AnnouncementRepository;
import pucrs.hslcommunicationservice.repository.UserAnnouncementViewRepository;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import java.io.IOException;
import java.time.LocalDateTime;
import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Collectors;

@Service
public class AnnouncementService {

    @Autowired
    private AnnouncementRepository announcementRepository;

    @PersistenceContext
    private EntityManager entityManager;

    @Autowired
    private UserService userService;

    @Autowired
    private S3Client s3Client;

    @Autowired
    private SNSClient snsClient;

    @Autowired
    private UserAnnouncementViewRepository viewRepository;

    public Announcement addAnnouncement(Announcement announcement) {
        List<Attachment> attachments = uploadAttachments(announcement);

        snsClient.publish(announcement.getContent(), announcement.getTitle());

        announcement.setAttachments(attachments);
        announcement.setDate(LocalDateTime.now());

        Announcement persisted = announcementRepository.save(announcement);

        userService.getUsers()
                .parallelStream()
                .forEach(user -> viewRepository.save(
                                UserAnnouncementView.builder()
                                        .user(user)
                                        .viewed(false)
                                        .announcement(announcement)
                                        .build()
                        )
                );

        return persisted;
    }

    public List<Announcement> getAnnouncements(Integer page, Integer pageSize, Long idDepartment, String token) {
        Long userRegistry = Long.parseLong(TokenAuthenticationService.getUserFromToken(token));

        List<Announcement> announcements = getAnnouncements(page, pageSize, idDepartment);
        viewRepository.findAll()
                .stream()
                .filter(view -> view.getUser().getRegistry().equals(userRegistry))
                .peek(view -> {
                    if (view.isViewed()) {
                        view.getAnnouncement().setViewed(true);
                    }
                });

    }

    private List<Announcement> getAnnouncements(Integer page, Integer pageSize, Long idDepartment) {
        if (page == null) page = 0;
        if (pageSize == null) pageSize = 30;
        if (idDepartment == null) {
            return announcementRepository.findAll(PageRequest.of(page, pageSize)).toList();
        } else {
            return entityManager
                    .createQuery("select a from Announcement a where a.department.id=:idDepartment", Announcement.class)
                    .setParameter("idDepartment", idDepartment)
                    .setFirstResult(page * pageSize)
                    .setMaxResults(pageSize)
                    .getResultList();
        }
    }

    public Announcement getAnnouncementById(Long id, String token) {
        Long userRegistry = Long.parseLong(TokenAuthenticationService.getUserFromToken(token));
        Announcement announcement = announcementRepository.getById(id);

        viewRepository.save(viewRepository.findAll()
                .stream()
                .filter(userView -> Objects.equals(userView.getUser().getRegistry(), userRegistry))
                .findFirst()
                .orElseThrow()
        );

        List<Attachment> attachments = announcement.getAttachments()
                .parallelStream()
                .peek(attachment -> {
                    String contentAsBase64 = null;
                    try {
                        contentAsBase64 = s3Client.getAttachmentAsBase64(attachment.getFileName());
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                    attachment.setContentAsBase64(contentAsBase64);
                })
                .collect(Collectors.toList());

        announcement.setAttachments(attachments);
        return announcement;
    }

    private String generateUniqueAttachmentFilename(String filename) {
        return "anexo-" + UUID.randomUUID() + "-" + filename;
    }

    public boolean deleteAnnnouncementById(Long announcementId) {
        try {
            announcementRepository.getById(announcementId).getAttachments()
                    .parallelStream()
                    .forEach(attach -> s3Client.removeAttachment(attach.getFileName()));
            announcementRepository.deleteById(announcementId);
            return true;
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    public Announcement editAnnouncement(Announcement announcement) {
        List<Attachment> attachments = uploadAttachments(announcement);
        List<Attachment> persistedAttachments = announcementRepository.getById(announcement.getId()).getAttachments();
        persistedAttachments.addAll(attachments);
        announcement.setAttachments(persistedAttachments);
        return announcementRepository.save(announcement);
    }

    private List<Attachment> uploadAttachments(Announcement announcement) {
        return announcement.getAttachments()
                .parallelStream()
                .peek(attachment -> attachment.setFileName(generateUniqueAttachmentFilename(attachment.getFileName())))
                .peek(attachment -> s3Client.uploadAttachment(attachment.getFileName(), attachment.getContentAsBase64()))
                .collect(Collectors.toList());
    }
}